@@ -1,28 +1,30 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+from collections import defaultdict, namedtuple
 from decimal import Decimal
-from collections import namedtuple, defaultdict
 from itertools import groupby
 
 from sql import Null
-from sql.conditionals import Coalesce
 from sql.aggregate import Max, Sum
+from sql.conditionals import Coalesce
+from sql.operators import Concat
 
 from trytond.config import config
 from trytond.i18n import gettext
-from trytond.model import Workflow, ModelView, ModelSQL, fields, Check, \
-    sequence_ordered, DictSchemaMixin
+from trytond.model import (
+    Check, DictSchemaMixin, ModelSQL, ModelView, Workflow, fields,
+    sequence_ordered)
 from trytond.model.exceptions import AccessError
-from trytond.pyson import Eval, If, Bool
-from trytond.transaction import Transaction
-from trytond.pool import Pool
-from trytond.rpc import RPC
-from trytond.wizard import Wizard, StateView, StateAction, Button
 from trytond.modules.company import CompanyReport
 from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool
+from trytond.pyson import Bool, Eval, If
+from trytond.rpc import RPC
+from trytond.transaction import Transaction
+from trytond.wizard import Button, StateAction, StateView, Wizard
 
 from .exceptions import (
-    StatementValidateError, StatementValidateWarning, StatementPostError)
+    StatementPostError, StatementValidateError, StatementValidateWarning)
 
 if config.getboolean('account_statement', 'filestore', default=False):
     file_id = 'origin_file_id'
@@ -51,30 +53,26 @@ class Statement(Workflow, ModelSQL, ModelView):
     __name__ = 'account.statement'
 
     _states = {'readonly': Eval('state') != 'draft'}
-    _depends = ['state']
     _balance_states = _states.copy()
     _balance_states.update({
             'invisible': ~Eval('validation', '').in_(['balance']),
             'required': Eval('validation', '').in_(['balance']),
             })
-    _balance_depends = _depends + ['validation']
     _amount_states = _states.copy()
     _amount_states.update({
             'invisible': ~Eval('validation', '').in_(['amount']),
             'required': Eval('validation', '').in_(['amount']),
             })
-    _amount_depends = _depends + ['validation']
     _number_states = _states.copy()
     _number_states.update({
             'invisible': ~Eval('validation', '').in_(['number_of_lines']),
             'required': Eval('validation', '').in_(['number_of_lines']),
             })
-    _number_depends = _depends + ['validation']
 
     name = fields.Char('Name', required=True)
     company = fields.Many2One(
         'company.company', "Company", required=True, select=True,
-        states=_states, depends=_depends)
+        states=_states)
     journal = fields.Many2One('account.statement.journal', 'Journal',
         required=True, select=True,
         domain=[
@@ -82,36 +80,33 @@ class Statement(Workflow, ModelSQL, ModelView):
             ],
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
-            },
-        depends=['state', 'company'])
+            })
     currency = fields.Function(fields.Many2One(
             'currency.currency', "Currency"), 'on_change_with_currency')
     date = fields.Date('Date', required=True, select=True)
     start_balance = Monetary(
         "Start Balance", currency='currency', digits='currency',
-        states=_balance_states, depends=_balance_depends)
+        states=_balance_states)
     end_balance = Monetary(
         "End Balance", currency='currency', digits='currency',
-        states=_balance_states, depends=_balance_depends)
+        states=_balance_states)
     balance = fields.Function(Monetary(
             "Balance", currency='currency', digits='currency',
-            states=_balance_states, depends=_balance_depends),
+            states=_balance_states),
         'on_change_with_balance')
     total_amount = Monetary(
         "Total Amount", currency='currency', digits='currency',
-        states=_amount_states, depends=_amount_depends)
+        states=_amount_states)
     number_of_lines = fields.Integer('Number of Lines',
-        states=_number_states, depends=_number_depends)
+        states=_number_states)
     lines = fields.One2Many('account.statement.line', 'statement',
         'Lines', states={
             'readonly': (Eval('state') != 'draft') | ~Eval('journal'),
-            },
-        depends=['state', 'journal'])
+            })
     origins = fields.One2Many('account.statement.origin', 'statement',
         "Origins", states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     origin_file = fields.Binary(
         "Origin File", readonly=True,
         file_id=file_id, store_prefix=store_prefix)
@@ -121,16 +116,16 @@ class Statement(Workflow, ModelSQL, ModelView):
             ('validated', "Validated"),
             ('cancelled', "Cancelled"),
             ('posted', "Posted"),
-            ], "State", readonly=True, select=True)
+            ], "State", readonly=True, select=True, sort=False)
     validation = fields.Function(fields.Char('Validation'),
         'on_change_with_validation')
     to_reconcile = fields.Function(
         fields.Boolean("To Reconcile"), 'get_to_reconcile')
 
-    del _states, _depends
-    del _balance_states, _balance_depends
-    del _amount_states, _amount_depends
-    del _number_states, _number_depends
+    del _states
+    del _balance_states
+    del _amount_states
+    del _number_states
 
     @classmethod
     def __setup__(cls):
@@ -229,12 +224,6 @@ class Statement(Workflow, ModelSQL, ModelView):
         if self.journal:
             return self.journal.currency.id
 
-    def get_end_balance(self, name):
-        end_balance = self.start_balance
-        for line in self.lines:
-            end_balance += line.amount
-        return end_balance
-
     @fields.depends('start_balance', 'end_balance')
     def on_change_with_balance(self, name=None):
         return ((getattr(self, 'end_balance', 0) or 0)
@@ -249,12 +238,12 @@ class Statement(Workflow, ModelSQL, ModelView):
 
         invoices = set()
         for line in self.lines:
-            if (getattr(line, 'invoice', None)
+            if (line.invoice
                     and line.invoice.currency == self.company.currency):
                 invoices.add(line.invoice)
         for origin in self.origins:
             for line in origin.lines:
-                if (getattr(line, 'invoice', None)
+                if (line.invoice
                         and line.invoice.currency == self.company.currency):
                     invoices.add(line.invoice)
         invoice_id2amount_to_pay = {}
@@ -269,7 +258,8 @@ class Statement(Workflow, ModelSQL, ModelView):
         for origin in origins:
             lines = list(origin.lines)
             for line in lines:
-                if (getattr(line, 'invoice', None) and line.id
+                if (line.invoice
+                        and line.id
                         and line.invoice.id in invoice_id2amount_to_pay):
                     amount_to_pay = invoice_id2amount_to_pay[line.invoice.id]
                     if (amount_to_pay
@@ -296,7 +286,7 @@ class Statement(Workflow, ModelSQL, ModelView):
 
         invoices = set()
         for line in self.lines:
-            if (getattr(line, 'invoice', None)
+            if (line.invoice
                     and line.invoice.currency == self.company.currency):
                 invoices.add(line.invoice)
         invoice_id2amount_to_pay = {}
@@ -310,7 +300,7 @@ class Statement(Workflow, ModelSQL, ModelView):
         lines = list(self.lines)
         line_offset = 0
         for index, line in enumerate(self.lines or []):
-            if getattr(line, 'invoice', None) and line.id:
+            if line.invoice and line.id:
                 if line.invoice.id not in invoice_id2amount_to_pay:
                     continue
                 amount_to_pay = invoice_id2amount_to_pay[line.invoice.id]
@@ -494,7 +484,7 @@ class Statement(Workflow, ModelSQL, ModelView):
                     gettext('account_statement'
                         '.msg_statement_invoice_paid_cancelled'))
             Line.write(paid_cancelled_invoice_lines, {
-                    'invoice': None,
+                    'related_to': None,
                     })
 
         cls.create_move(statements)
@@ -503,9 +493,10 @@ class Statement(Workflow, ModelSQL, ModelView):
                 'state': 'validated',
                 })
         common_lines = [l for l in Line.search([
-                ('statement.state', '=', 'draft'),
-                ('invoice.state', 'in', ['posted', 'paid']),
-                ])
+                    ('statement.state', '=', 'draft'),
+                    ('related_to.state', 'in', ['posted', 'paid'],
+                        'account.invoice'),
+                    ])
             if l.invoice.reconciled]
         if common_lines:
             warning_key = '_'.join(str(l.id) for l in common_lines)
@@ -514,7 +505,7 @@ class Statement(Workflow, ModelSQL, ModelView):
                     gettext('account_statement'
                         '.msg_statement_paid_invoice_draft'))
             Line.write(common_lines, {
-                    'invoice': None,
+                    'related_to': None,
                     })
 
     @classmethod
@@ -665,13 +656,12 @@ class Statement(Workflow, ModelSQL, ModelView):
         return new_statements
 
 
-def origin_mixin(_states, _depends):
+def origin_mixin(_states):
     class Mixin:
         __slots__ = ()
         statement = fields.Many2One(
             'account.statement', "Statement",
-            required=True, ondelete='CASCADE', states=_states,
-            depends=_depends)
+            required=True, ondelete='CASCADE', states=_states)
         statement_state = fields.Function(
             fields.Selection('get_statement_states', "Statement State"),
             'on_change_with_statement_state')
@@ -680,10 +670,10 @@ def origin_mixin(_states, _depends):
             'on_change_with_company', searcher='search_company')
         number = fields.Char("Number")
         date = fields.Date(
-            "Date", required=True, states=_states, depends=_depends)
+            "Date", required=True, states=_states)
         amount = Monetary(
             "Amount", currency='currency', digits='currency', required=True,
-            states=_states, depends=_depends)
+            states=_states)
         currency = fields.Function(fields.Many2One(
                 'currency.currency', "Currency"), 'on_change_with_currency')
         party = fields.Many2One(
@@ -691,7 +681,7 @@ def origin_mixin(_states, _depends):
             context={
                 'company': Eval('company', -1),
                 },
-            depends=_depends + ['company'])
+            depends={'company'})
         account = fields.Many2One(
             'account.account', "Account",
             domain=[
@@ -702,9 +692,8 @@ def origin_mixin(_states, _depends):
             context={
                 'date': Eval('date'),
                 },
-            states=_states, depends=_depends + ['company', 'date'])
-        description = fields.Char(
-            "Description", states=_states, depends=_depends)
+            states=_states, depends={'date'})
+        description = fields.Char("Description", states=_states)
 
         @classmethod
         def __setup__(cls):
@@ -742,12 +731,9 @@ def origin_mixin(_states, _depends):
 _states = {
     'readonly': Eval('statement_state') != 'draft',
     }
-_depends = ['statement_state']
 
 
-class Line(
-        origin_mixin(_states, _depends), sequence_ordered(),
-        ModelSQL, ModelView):
+class Line(origin_mixin(_states), sequence_ordered(), ModelSQL, ModelView):
     'Account Statement Line'
     __name__ = 'account.statement.line'
 
@@ -756,18 +742,24 @@ class Line(
             ('company', '=', Eval('company', -1)),
             ],
         depends=['company'])
-    invoice = fields.Many2One('account.invoice', 'Invoice',
-        domain=[
-            ('company', '=', Eval('company', -1)),
-            If(Bool(Eval('party')), [('party', '=', Eval('party'))], []),
-            If(Bool(Eval('account')), [('account', '=', Eval('account'))], []),
-            If(Eval('statement_state') == 'draft',
-                ('state', '=', 'posted'),
-                ('state', '!=', '')),
-            ],
+    related_to = fields.Reference(
+        "Related To", 'get_relations',
+        domain={
+            'account.invoice': [
+                ('company', '=', Eval('company', -1)),
+                If(Bool(Eval('party')),
+                    ('party', '=', Eval('party')),
+                    ()),
+                If(Bool(Eval('account')),
+                    ('account', '=', Eval('account')),
+                    ()),
+                If(Eval('statement_state') == 'draft',
+                    ('state', '=', 'posted'),
+                    ('state', '!=', '')),
+                ],
+            },
         states=_states,
-        context={'with_payment': False},
-        depends=['company', 'party', 'account'] + _depends)
+        context={'with_payment': False})
     origin = fields.Many2One('account.statement.origin', 'Origin',
         readonly=True,
         states={
@@ -776,18 +768,16 @@ class Line(
         domain=[
             ('statement', '=', Eval('statement', -1)),
             ('date', '=', Eval('date', None)),
-            ],
-        depends=['statement', 'date'])
+            ])
 
     @classmethod
     def __setup__(cls):
         super(Line, cls).__setup__()
-        if 'origin' not in cls.date.depends:
-            cls.date.states.update({
-                    'readonly': (cls.date.states['readonly']
-                        | Bool(Eval('origin', 0))),
-                    })
-            cls.date.depends.append('origin')
+        cls.date.states = {
+            'readonly': (
+                (Eval('statement_state') != 'draft')
+                | Bool(Eval('origin', 0))),
+            }
         cls.account.required = True
         t = cls.__table__()
         cls._sql_constraints += [
@@ -795,11 +785,51 @@ class Line(
                 'account_statement.msg_line_amount_non_zero'),
             ]
 
+    @classmethod
+    def __register__(cls, module):
+        table = cls.__table__()
+
+        super().__register__(module)
+
+        table_h = cls.__table_handler__(module)
+        cursor = Transaction().connection.cursor()
+        if table_h.column_exist('invoice'):
+            cursor.execute(*table.update(
+                    [table.related_to],
+                    [Concat('account.invoice,', table.invoice)],
+                    where=table.invoice != Null))
+            table_h.drop_column('invoice')
+
+    @property
+    @fields.depends('related_to')
+    def invoice(self):
+        pool = Pool()
+        Invoice = pool.get('account.invoice')
+        related_to = getattr(self, 'related_to', None)
+        if isinstance(related_to, Invoice) and related_to.id >= 0:
+            return related_to
+
+    @invoice.setter
+    def invoice(self, value):
+        self.related_to = value
+
     @staticmethod
     def default_amount():
         return Decimal(0)
 
-    @fields.depends('amount', 'party', 'invoice', 'date')
+    @classmethod
+    def _get_relations(cls):
+        "Return a list of Model names for related_to Reference"
+        return ['account.invoice']
+
+    @classmethod
+    def get_relations(cls):
+        Model = Pool().get('ir.model')
+        get_name = Model.get_name
+        models = cls._get_relations()
+        return [(None, '')] + [(m, get_name(m)) for m in models]
+
+    @fields.depends('amount', 'party', 'date', methods=['invoice'])
     def on_change_party(self):
         if self.party:
             if self.amount:
@@ -816,8 +846,10 @@ class Line(
             else:
                 self.invoice = None
 
-    @fields.depends('amount', 'party', 'account', 'invoice', 'statement',
-        'date', '_parent_statement.journal')
+    @fields.depends(
+        'amount', 'party', 'account', 'date',
+        'statement', '_parent_statement.journal',
+        methods=['invoice'])
     def on_change_amount(self):
         Currency = Pool().get('currency.currency')
         if self.party:
@@ -844,7 +876,7 @@ class Line(
             else:
                 self.invoice = None
 
-    @fields.depends('account', 'invoice')
+    @fields.depends('account', methods=['invoice'])
     def on_change_account(self):
         if self.invoice:
             if self.account:
@@ -853,8 +885,8 @@ class Line(
             else:
                 self.invoice = None
 
-    @fields.depends('party', 'account', 'invoice')
-    def on_change_invoice(self):
+    @fields.depends('party', 'account', methods=['invoice'])
+    def on_change_related_to(self):
         if self.invoice:
             if not self.party:
                 self.party = self.invoice.party
@@ -914,7 +946,7 @@ class Line(
         else:
             default = default.copy()
         default.setdefault('move', None)
-        default.setdefault('invoice', None)
+        default.setdefault('related_to', None)
         return super(Line, cls).copy(lines, default=default)
 
     @classmethod
@@ -999,7 +1031,7 @@ class Line(
             )
 
 
-del _states, _depends
+del _states
 
 
 class LineGroup(ModelSQL, ModelView):
@@ -1055,7 +1087,7 @@ class LineGroup(ModelSQL, ModelView):
         return move.join(line,
             condition=move.id == line.move
             ).select(*columns,
-                where=move.origin.ilike(Statement.__name__ + ',%'),
+                where=move.origin.like(Statement.__name__ + ',%'),
                 group_by=std_columns + [move.id]
                 )
 
@@ -1073,10 +1105,9 @@ class LineGroup(ModelSQL, ModelView):
 _states = {
     'readonly': (Eval('statement_state') != 'draft') | Eval('lines', []),
     }
-_depends = ['statement_state']
 
 
-class Origin(origin_mixin(_states, _depends), ModelSQL, ModelView):
+class Origin(origin_mixin(_states), ModelSQL, ModelView):
     "Account Statement Origin"
     __name__ = 'account.statement.origin'
     _rec_name = 'number'
@@ -1090,8 +1121,7 @@ class Origin(origin_mixin(_states, _depends), ModelSQL, ModelView):
         domain=[
             ('statement', '=', Eval('statement', -1)),
             ('date', '=', Eval('date', None)),
-            ],
-        depends=['statement', 'date', 'statement_id'])
+            ])
     statement_id = fields.Function(
         fields.Integer("Statement ID"), 'on_change_with_statement_id')
     pending_amount = fields.Function(Monetary(
@@ -1145,7 +1175,7 @@ class Origin(origin_mixin(_states, _depends), ModelSQL, ModelView):
         return super().copy(origins, default=default)
 
 
-del _states, _depends
+del _states
 
 
 class OriginInformation(DictSchemaMixin, ModelSQL, ModelView):
